/*
 * File access convenience functions.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>
#include <systypes/fcntl.h>
#include <systypes/stat.h>

#include <stdio.h>
#include <string.h>

#include "dataspace_client.h"
#include "directory_client.h"
#include "file_client.h"
#include "filesystem_client.h"
#include "flush_client.h"
#include "opener_client.h"
#include "opener_context_client.h"
#include "pipe_client.h"
#include "pipe_opener_client.h"
#include "mapped_file_client.h"

#include "file.h"
#include "notifier.h"



/* Update the extent of the file in a region using the region start and end
   positions and the file size. */

static void _update_extent(file_t *file)
{
  /* Handle files ending after or within the region. */

  if (file->size > file->start_pos)
  {
    if (file->size > file->end_pos)
      file->data_end = file->end_pos - file->start_pos;
    else
      file->data_end = file->size - file->start_pos;
  }

  /* Handle files ending before the region. */

  else
    file->data_end = 0;
}



/* Obtain a reference to a file opener for the given user. */

long file_open_for_user(user_t user, l4_cap_idx_t server, l4_cap_idx_t *opener)
{
  client_Filesystem filesystem(server);

  return filesystem.open_for_user(user, opener);
}



/* Initialise the given file structure. */

void file_init(file_t *file)
{
  file->memory = NULL;
  file->ref = L4_INVALID_CAP;
  file->start_pos = 0;
  file->end_pos = 0;
  file->data_end = 0;
  file->data_current = 0;
  file->object_flags = 0;
  file->can_block = 0;
  file->notifications = 0;
  file->flags = 0;
  file->error = 0;
}



/* Release resources for the given file. */

void file_close(file_t *file)
{
  if (l4_is_valid_cap(file->ref))
  {
    ipc_cap_free_um(file->ref);

    /* Only unsubscribe after actually closing the file and sending any
       notifications. */

    notifier_get_task_notifier()->unsubscribe(file);
  }

  if (file->memory != NULL)
    ipc_detach_dataspace(file->memory);

  file_init(file);
}

/* Make a directory in the filesystem. This is a convenience function invoking
   file_context and file_context_mkdir. */

long file_mkdir(const char *filename, mode_t mode, l4_cap_idx_t server)
{
  file_t context;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, filename, 0, NULL))
    return -L4_ENOMEM;

  err = file_context_mkdir(systypes_to_sys_mode(mode), &context);

  /* Close the context, although a separate mechanism could permit contexts to
     remove several files. */

  file_close(&context);
  return err;
}

/* Open a file using the given structure, indicating the filename and
   filesystem server. The file_mmap function should be used to obtain access to
   memory providing file data. This is a convenience function invoking
   file_context and file_context_open. */

long file_open(file_t *file, const char *filename, flags_t flags, l4_cap_idx_t server)
{
  file_t context;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, filename, 0, NULL))
    return -L4_ENOMEM;

  err = file_context_open(file, flags, &context);

  /* Close the context, although a separate mechanism could permit contexts to
     open several files. */

  file_close(&context);
  return err;
}

/* Remove a file from the filesystem. This is a convenience function invoking
   file_context and file_context_remove. */

long file_remove(const char *filename, l4_cap_idx_t server)
{
  file_t context;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, filename, 0, NULL))
    return -L4_ENOMEM;

  err = file_context_remove(&context);

  /* Close the context, although a separate mechanism could permit contexts to
     remove several files. */

  file_close(&context);
  return err;
}

/* Rename an object in the filesystem. This is a convenience function invoking
   file_context and file_context_rename. */

long file_rename(const char *source, const char *target, l4_cap_idx_t server)
{
  file_t context;
  offset_t written;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, source, 0, &written))
    return -L4_ENOMEM;

  if (!file_string_set(&context, target, written + 1, NULL))
    return -L4_ENOMEM;

  err = file_context_rename(&context);

  /* Close the context, although a separate mechanism could permit contexts to
     rename several files. */

  file_close(&context);
  return err;
}

/* Obtain filesystem object statistics. This is a convenience function invoking
   file_context and file_context_stat. */

long file_stat(const char *filename, struct stat *st, l4_cap_idx_t server)
{
  file_t context;
  offset_t written;
  long err;

  err = file_context(&context, server);
  if (err)
    return err;

  if (!file_string_set(&context, filename, 0, &written))
    return -L4_ENOMEM;

  err = file_context_stat(st, &context);

  /* Close the context, although a separate mechanism could permit contexts to
     rename several files. */

  file_close(&context);
  return err;
}



/* Initialise a file structure for a context obtained from the given server
   attaching memory to communicate filename information. */

long file_context(file_t *file, l4_cap_idx_t server)
{
  if (l4_is_invalid_cap(server))
    return -L4_EINVAL;

  client_Opener opener(server);
  offset_t size;
  map_flags_t flags;
  long err;

  file_init(file);
  file->flags = O_RDWR;

  err = opener.context(&file->ref);
  if (err)
    return err;

  client_Dataspace context_ds(file->ref);

  err = context_ds.info(&size, &flags);
  if (err)
    return err;

  file->start_pos = 0;
  file->end_pos = size;

  return ipc_attach_dataspace(file->ref, size, (void **) &file->memory);
}

/* Make a directory using the given context. */

long file_context_mkdir(mode_t mode, file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  return openercontext.mkdir(mode);
}

/* Open a file using the given structure and context. */

long file_context_open(file_t *file, flags_t flags, file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  file_init(file);
  file->flags = flags;
  return openercontext.open(flags, &file->size, &file->ref, &file->object_flags);
}

/* Remove a file using the given context. */

long file_context_remove(file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  return openercontext.remove();
}

/* Rename a file using the given context. */

long file_context_rename(file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  return openercontext.rename();
}

/* Obtain filesystem object statistics using the given context. */

long file_context_stat(struct stat *st, file_t *context)
{
  client_OpenerContext openercontext(context->ref);
  long err = openercontext.stat();

  if (err)
    return err;

  /* Copy the stat structure manually in case of any divergence. */

  systypes_copy_from_sys_stat(st, (sys_stat_t *) context->memory);

  return L4_EOK;
}



/* Flush populated data and obtain an updated file size and populated data
   details. */

long file_flush(file_t *file)
{
  if (l4_is_invalid_cap(file->ref))
    return -L4_EINVAL;

  client_Flush _file(file->ref);
  long err = _file.flush(file->data_current, &file->size);

  if (err)
    return err;

  _update_extent(file);

  return L4_EOK;
}

/* Map a region of the given file to a memory region, obtaining an updated file
   size and populated data details. Unmap any previously mapped region. */

long file_mmap(file_t *file, offset_t position, offset_t length,
               offset_t start_visible, offset_t end_visible,
               l4re_rm_flags_t region_flags)
{
  long err = file_mmap_only(file, position, length, start_visible, end_visible);

  if (err)
    return err;

  if (file->memory != NULL)
    ipc_detach_dataspace(file->memory);

  file->memory = NULL;

  err = ipc_attach_dataspace_align(file->ref, file_span(file),
                                   L4RE_RM_F_SEARCH_ADDR | region_flags,
                                   L4_PAGESHIFT,
                                   (void **) &file->memory);
  if (err)
    return err;

  return L4_EOK;
}

/* Request access to a region of the given file, obtaining an updated file size
   and populated data details. The region is not mapped, however. */

long file_mmap_only(file_t *file, offset_t position, offset_t length,
                    offset_t start_visible, offset_t end_visible)
{
  client_MappedFile mapped_file(file->ref);
  long err = mapped_file.mmap(position, length, start_visible, end_visible,
                              &file->start_pos, &file->end_pos, &file->size);

  if (err)
    return err;

  _update_extent(file);

  return L4_EOK;
}

/* Return opening flags compatible with the given region flags. */

flags_t file_opening_flags(l4re_rm_flags_t rm_flags)
{
  if ((rm_flags & L4RE_RM_F_RW) == L4RE_RM_F_RW)
    return O_RDWR;
  else if (rm_flags & L4RE_RM_F_W)
    return O_WRONLY;
  else
    return O_RDONLY;
}

/* Return mmap flags corresponding to the file access flags. */

l4re_rm_flags_t file_region_flags(flags_t flags)
{
  l4re_rm_flags_t rm_flags;

  switch (flags & 3)
  {
    case O_WRONLY:
      rm_flags = L4RE_RM_F_W;
      break;

    case O_RDWR:
      rm_flags = L4RE_RM_F_RW;
      break;

    default:
      rm_flags = L4RE_RM_F_R;
      break;
  }

  return rm_flags;
}

/* Resize a file, obtaining updated file size and populated data details. */

long file_resize(file_t *file, offset_t size)
{
  if (!(file->object_flags & OBJECT_HAS_SIZE))
    return -L4_EIO;

  client_File _file(file->ref);
  offset_t file_size = size;
  long err = _file.resize(&file_size);

  if (err)
    return err;

  file->size = file_size;
  _update_extent(file);
  return L4_EOK;
}



/* Return the amount of data in the mapped region for the given file. */

offset_t file_populated_span(file_t *file)
{
  offset_t size = file_span(file);
  return (file->data_end < size) ? file->data_end : size;
}

/* Return the size of the mapped region for the given file. */

offset_t file_span(file_t *file)
{
  return file->end_pos - file->start_pos;
}



/* Get a pointer to any terminated string at the given offset or NULL if the
   data from offset is not terminated. */

char *file_string_get(file_t *file, offset_t offset)
{
  offset_t limit = file_span(file) - offset;

  if (strnlen(file->memory + offset, limit) < limit)
    return file->memory + offset;
  else
    return NULL;
}

/* Copy a string to the mapped region at the given offset, returning 1 (true)
   where all characters were copied, 0 (false) otherwise. The precise number of
   characters copied, excluding the zero terminator is provided via the written
   parameter if it is not specified as NULL. */

int file_string_set(file_t *file, const char *data, offset_t offset,
                    offset_t *written)
{
  offset_t i, pos, limit = file_span(file);

  /* Do not attempt to copy data with an invalid offset. */

  if (offset >= limit)
  {
    if (written != NULL)
      *written = 0;
    return 0;
  }

  /* Copy the data to the given offset, stopping at the end of the region. */

  for (i = 0, pos = offset; pos < limit; i++, pos++)
  {
    file->memory[pos] = data[i];

    /* Terminator written, can return immediately. */

    if (!data[i])
    {
      if (written != NULL)
        *written = pos - offset;
      return 1;
    }
  }

  /* Terminate the incomplete string at the end of the region. */

  file->memory[limit - 1] = '\0';
  if (written != NULL)
    *written = limit - 1 - offset;
  return 0;
}



/* Return the number of remaining populated bytes in the region. */

offset_t file_data_available(file_t *file)
{
  return file_populated_span(file) - file->data_current;
}

/* Return the current data offset in the region. */

char *file_data_current(file_t *file)
{
  return file->memory + file->data_current;
}

/* Return the current access position in the file. */

offset_t file_data_current_position(file_t *file)
{
  return file->start_pos + file->data_current;
}

/* Return the position of the end of the populated bytes in the region. */

offset_t file_data_end_position(file_t *file)
{
  return file->start_pos + file->data_end;
}

/* Return the amount of remaining space in the region. */

offset_t file_data_space(file_t *file)
{
  return file_span(file) - file->data_current;
}



/* Copy data to the given buffer from the current data position, updating the
   position. */

void file_data_read(file_t *file, char *buf, offset_t to_transfer)
{
  memcpy(buf, file_data_current(file), to_transfer);

  /* Update position details. */

  file->data_current += to_transfer;
}

/* Copy data from the given buffer to the current data position, updating the
   position and the extent of populated data if this was exceeded. */

void file_data_write(file_t *file, char *buf, offset_t to_transfer)
{
  memcpy(file_data_current(file), buf, to_transfer);

  /* Update position details. */

  file->data_current += to_transfer;

  if (file->data_current > file->data_end)
    file->data_end = file->data_current;
}



/* Opaque notifier type for file_notifier_t. */

struct file_notifier
{
  FileNotifier *obj;
};



/* Close a notifier object. */

void file_notify_close(file_notifier_t *notifier)
{
  delete notifier->obj;
  delete notifier;
}

/* Obtain a local notifier object. */

file_notifier_t *file_notify_local()
{
  file_notifier_t *notifier = new file_notifier_t;

  notifier->obj = notifier_get_local_notifier();
  return notifier;
}

/* Obtain the task-wide notifier object. */

file_notifier_t *file_notify_task()
{
  file_notifier_t *notifier = new file_notifier_t;

  notifier->obj = notifier_get_task_notifier();
  return notifier;
}

/* Subscribe to notification events on a file. */

long file_notify_subscribe(file_t *file, notify_flags_t flags, file_notifier_t *notifier)
{
  return notifier->obj->subscribe(file, flags);
}

/* Unsubscribe from notification events on a file. */

long file_notify_unsubscribe(file_t *file, file_notifier_t *notifier)
{
  return notifier->obj->unsubscribe(file);
}

/* Wait for a notification event on a file. */

long file_notify_wait_file(file_t *file, file_notifier_t *notifier)
{
  SpecificFileNotifier *specific_notifier = dynamic_cast<SpecificFileNotifier *>(notifier->obj);
  long err = specific_notifier->wait_file(file);

  /* Unsubscribe if a closure notification has been received. */

  if (!err && (file->notifications & NOTIFY_PEER_CLOSED))
    file_notify_unsubscribe(file, notifier);

  return err;
}

/* Wait for notification events on files. */

long file_notify_wait_files(file_t **file, file_notifier_t *notifier)
{
  GeneralFileNotifier *general_notifier = dynamic_cast<GeneralFileNotifier *>(notifier->obj);
  long err = general_notifier->wait(file);

  /* Unsubscribe if a closure notification has been received. */

  if (!err && ((*file)->notifications & NOTIFY_PEER_CLOSED))
    file_notify_unsubscribe(*file, notifier);

  return err;
}



/* Open two pipe endpoints using the given pipe server. */

long pipe_open(offset_t size, file_t *reader, file_t *writer, l4_cap_idx_t server)
{
  if (l4_is_invalid_cap(server))
    return -L4_EINVAL;

  client_PipeOpener opener(server);

  file_init(reader);
  file_init(writer);
  reader->flags = O_RDONLY;
  writer->flags = O_WRONLY;

  return opener.pipe(size, &reader->ref, &writer->ref);
}

/* Access the current region for a pipe endpoint. */

long pipe_current(file_t *pipe)
{
  client_Pipe _pipe(pipe->ref);
  long err = _pipe.current_region(&pipe->data_end, &pipe->size);

  if (err)
    return err;

  pipe->end_pos = pipe->size;

  /* Attach memory if necessary. */

  if (pipe->memory == NULL)
  {
    err = ipc_attach_dataspace(pipe->ref, file_span(pipe), (void **) &pipe->memory);
    if (err)
      return err;
  }

  return L4_EOK;
}

/* Access the next region for a pipe endpoint, updating the eventual size of
   the current region. */

long pipe_next(file_t *pipe)
{
  client_Pipe _pipe(pipe->ref);
  long err = _pipe.next_region(&pipe->data_end, &pipe->size);

  if (err)
    return err;

  if (pipe->memory != NULL)
    ipc_detach_dataspace(pipe->memory);

  pipe->end_pos = pipe->size;
  pipe->memory = NULL;

  err = ipc_attach_dataspace(pipe->ref, file_span(pipe), (void **) &pipe->memory);
  if (err)
    return err;

  return L4_EOK;
}

/* Set the size of the written region. */

long pipe_written(file_t *pipe, offset_t size)
{
  if (size <= pipe->size)
  {
    pipe->data_end = size;
    return L4_EOK;
  }
  else
    return -L4_EINVAL;
}



/* Obtain a directory listing stream from a directory. */

long directory_opendir(file_t *file, file_t *reader)
{
  client_Directory directory(file->ref);
  file_init(reader);
  reader->flags = O_RDONLY;
  return directory.opendir(&reader->size, &reader->ref, &reader->object_flags);
}

// vim: tabstop=2 expandtab shiftwidth=2
